use std::path::PathBuf;

use axum::{extract::State, response::Html, routing::{get, post}, Form, Router};
use pandoc::{InputFormat, InputKind, MarkdownExtension, OutputKind};
use serde::{Deserialize, Serialize};
use sqlx::SqlitePool;
use strum::EnumIter;
use strum::IntoEnumIterator;
use tera::{Context, Tera};

use tower_http::services::ServeDir; 

use tracing::{debug, error, Level};

#[derive(Clone)]
struct AppState {
    tera: Tera,
    pool: SqlitePool
}

impl AppState {
    async fn new() -> Self {
        let pool = sqlx::sqlite::SqlitePool::connect(DB_URL).await.unwrap();

        let tera = match Tera::new("./templates/**/*.html") {
            Ok(t) => {
                for tpl in t.get_template_names() {
                    debug!("Added template: {tpl}");
                }
                debug!("Template loaded");
                t
            }
            Err(e) => {
                error!("Parsing errors: {e}");
                println!("Parsing errors: {e}");
                ::std::process::exit(1);
            }
        };

        AppState {
            pool,
            tera,
        }
    }
}

const DB_URL: &str = "sqlite://sqlite.db";

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt().with_max_level(Level::DEBUG).init();

    let state = AppState::new().await;

    // build our application with a single route
    let app = Router::new()
        .route("/", get(inscription))
        .route("/inscription", get(inscription))
        .route("/recapitulation", post(recapitulation))
        // Permet de servir les fichiers PDF des dossiers générés (static file)
        // TODO: mettre une fallback
        .nest_service("/dossiers", ServeDir::new("./dossiers"))
        // .nest("dossiers", get(get_static_file))
        .with_state(state);

    // run our app with hyper, listening globally on port 3000
    let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await.unwrap();
    axum::serve(listener, app).await.unwrap();
}

async fn inscription(State(state): State<AppState>) -> Html<String> {
    let teracontext = tera::Context::new();
    Html(state.tera.render(Template::MainForm.display(), &teracontext).unwrap())
}

#[derive(Serialize, Deserialize, Debug, Clone, sqlx::FromRow)]
struct Candidate {
    #[sqlx(rename="prenom")] 
    firstname: String,
    #[sqlx(rename="nom")] 
    lastname: String,
}
impl PartialEq for Candidate {
    fn eq(&self, other: &Self) -> bool {
        self.firstname == other.firstname && self.lastname == other.lastname
    }
}

#[derive(EnumIter)]
enum Template {
    MainForm,
    Pdf
}
impl Template {
    fn display(&self) -> &str {
        match self {
            Template::Pdf => "pdf.html",
            Template::MainForm => "index.html"
        }
    }
}

async fn recapitulation(State(state): State<AppState>, Form(candidate): Form<Candidate>) -> Html<String> {
    let inserted_candidate = db_add_candidate(&state, candidate).await.unwrap();

    debug!("PDF candidate: {:?}", inserted_candidate);

    let rendered_data = state.tera.render(Template::Pdf.display(), &Context::from_serialize(&inserted_candidate).unwrap()).unwrap();

    let mut pandoc = pandoc::new();
    pandoc.set_input(InputKind::Pipe(rendered_data));
    pandoc.set_input_format(InputFormat::Html,[MarkdownExtension::RawHtml].to_vec());
    pandoc.set_output(OutputKind::File(PathBuf::from("tmp.pdf")));
    pandoc.set_show_cmdline(true);
    let _ = pandoc.execute().unwrap();

    Html(state.tera.render("dossier.html", &Context::from_serialize(&inserted_candidate).unwrap()).unwrap())
}

async fn db_add_candidate(state: &AppState, candidate: Candidate) -> Result<Candidate, sqlx::Error>{
    debug!("db_add_candidate: {} {}", candidate.firstname, candidate.lastname);
    let insert = sqlx::query("insert into users (prenom, nom) values ($1,$2)")
        .bind(&candidate.firstname)
        .bind(&candidate.lastname)
        .execute(&state.pool).await;
    match insert {
        Ok(r) => {
            let userid = r.last_insert_rowid();
            sqlx::query_as::<_, Candidate>("select prenom, nom from users where id = $1")
                .bind(userid)
                .fetch_one(&state.pool).await
        }
        Err(e) => {
            error!("add_candidate error: {}", e);
            Err(e)
        }
    }
}

#[cfg(test)]
#[tokio::test]
async fn test_sql_candidate() {
    let state = AppState::new().await;
    let candidate = Candidate {
        lastname: "Fery".to_string(),
        firstname: "Yann".to_string()
    };
    let inserted_candidate = db_add_candidate(&state, candidate.clone()).await.unwrap();
    assert_eq!(candidate, inserted_candidate);

}

#[tokio::test]
async fn test_template_list() {
    let state = AppState::new().await;

    let mut tera: Vec<&str> = [].to_vec();
    for tpl in state.tera.get_template_names() {
        tera.push(tpl);
    }
    let mut list: Vec<String> = [].to_vec();
    for tpl in Template::iter() {
        list.push(tpl.display().to_string());
    }
    assert_eq!(tera, list);
}
